import socket, select, sys, regex, time, regsub, string, types
import data, player, mud

##################################################################
# Client class
# high level application kinda thingy
##################################################################
class client:

    def __init__(self):
	data.timeout = .1
	data.theapp = self

    # the main loop of lyntin
    # gets input, sends it to the mud
    # gets output, shows it to the player
    # catches exceptions
    def Mainloop(self):
	while 1:
	    try:
		# get input for all connected sessions
		for ses in data.sessionlist:
		    if ses.connected:
			datafrom = ses.ReadMud()
			# handle actions and displaying stuff from mud
			mud.HandleMudOutput(datafrom, ses)
		    datato = self.GetUserInput()
		    if datato == '\n':
			self.SendPlainInput('\r')
		    else:
			self.PreHandleUserInput(datato)
	    except KeyboardInterrupt:
		player.Quit()
	    except SystemExit: # minor kludge to handle sys.exit
		break
	    # if anything else goes wrong, we get ugly and print a traceback
 	    except:
		player.Putline('ack! error!')
		import traceback
		traceback.print_exc()

    # process the command line
    def CommandLine(self):
	if len(sys.argv) > 1:
	    for n in sys.argv[1:]:
		player.ParseFile(['load', n])
	pass

    # check for stuff from stdin
    def GetUserInput(self):
	readers,w,e = select.select([sys.stdin], [], [], data.timeout)
	if not readers: # timer expired
	    return  
	for R in readers:
	    thedata = R.readline()
	    return thedata

    # do stuff that we want to do one time for each command, like
    # registering the command in the history list.
    # we can't do this in HandleUserInput because it is recursive
    def PreHandleUserInput(self, input):
	if input:
	    self.RecordHistory(input)
	    self.HandleUserInput(input)

    # deal with input from user
    # regular text gets worked over for aliases, then sent to mud
    # meta-commands (#[0-9], #sesname command, etc) handled here
    # this function is recursive
    def HandleUserInput(self, input):
	if not input:
	    return
	ses = None
	mud.log('HandleUserInput: received ' + input)
	# trim leading/trailing whitespace
	input = string.strip(input)

	# check for a sequence of commands separated by ';'
	whether, result = IsSequence(input)
	if whether:
	    for s in result:
		self.HandleUserInput(s)
	    return
	
	# IsSequence() can modify input
	input = result

	# check for a braced command
	if IsBrace(input):
	    input = input[1:-1]
	    self.HandleUserInput(input)
	    return

	# fill in values for any variables
	whether, input = self.SubVars(input)
	if whether:
	    self.HandleUserInput(input)
	    return

	# check for a reference to the history list
	if input[0] == '!':
	    self.DoHistoricCommand(input)
	    return

	# work input over for aliases/speedwalking
	whether, input = player.WorkOver(input)

	if whether == 'speed':
	    # speedwalk
	    self.DoSpeedWalk(input)
	    return

	# triggered an alias, recurse
	elif whether:
	    self.HandleUserInput(input)
	    return

	# handle commands to client
	if input and input[0] == '#':
	    if input[-1] == '\n':
		input = input[1:-1] # remove '#' at beginning and '\n' at end
	    else: input = input[1:]
	    words = string.split(input)

	    # see if the command applies to a certain session
	    ses = data.GetSes(words[0])
	    if ses:
		 self.SessionCommand(words, ses)
		 return

	    # repeat command?
	    if self.RepeatCommand(words):
		return 

	    # ok, just apply a command to the current session
	    player.DoCommand(input)
	    return
	
	# expanded to non-command, just send it
	return self.SendPlainInput(input)

    # redo a command from the history list
    def DoHistoricCommand(self, input):
	if regex.search('^![0-9]*', input) == -1:
	    # not a valid history command
	    self.SendPlainInput(input)
	else:
	    num = HistNumber(input)
	    # we need to add one to num because the history list
	    # has already increased in size thanks to PreHandleUserInput
	    h = data.history[num+1]
	    input = self.DoHistorySubs(input, h)
	    self.HandleUserInput(input)

    # fill in values for any variables found in the input
    # unless they're braced
    def SubVars(self, input):
	if string.find(input, '$') == -1:
	    # no variables in this
	    return 0, input
	nesting = 0
	varhunt = 0
	parsed = ''
	v = ''
	whether = 0
	for c in input:
	    if varhunt:
		if data.currsession.Isvar(v+c):
		    v = v + c
		elif data.currsession.IsRealvar(v):
		    # found a complete variable
		    parsed = parsed[:-1] + data.currsession.vars[v] # chop off '$' from parsed
		    parsed = parsed + c
		    whether = 1
		    varhunt = 0
		    v = ''
		else:
		    # this is not a variable. just put v back onto parsed
		    parsed = parsed + v
		    v = ''
		    varhunt = 0
	    elif c == '{':
		parsed = parsed + c
		nesting = nesting + 1
	    elif c == '}':
		parsed = parsed + c
		nesting = nesting - 1
	    elif c == '$':
		parsed = parsed + c
		if not nesting:
		    varhunt = 1
	    else:
		parsed = parsed + c
	if v:
	    # there could've been a variable at the end of the input
	    for var in data.currsession.vars.keys():
		if v == var:
		    whether = 1
		    parsed = parsed[:-1] + data.currsession.vars[v]
	return whether, parsed

    # save input in the history list
    def RecordHistory(self, input):
	if input == '\n':
	    # don't record this crap
	    return
	if IsHistory(input):
	    # we want to record the actual command, instead of 
	    # something like !4, so we have to look it up
	    num = HistNumber(input)
	    old = data.history[num]
	    # perform any requested substitutions
	    input = self.DoHistorySubs(input, old)

	# do what we came here for
	data.history = [input] + data.history
	# 'allocate' a new history list if this one 
	# has gotten too huge
	if len(data.history) >= 2 * data.histsize:
	    newhist = []
	    for i in range(data.histsize):
		newhist = newhist + [data.history[i]]
	    data.history = newhist

    # when requesting a redo from the history list, the user has
    # the option of substituting for some of the text in the command, 
    # along the lines of !3 jay=joe.
    # here's where we do that.
    def DoHistorySubs(self, hist, input):
	# looks like '!4 joe=john' or '! I like=I hate'
	sub_regex = regex.compile('^![0-9]* \(.*\)=\(.*\)')
	if sub_regex.search(hist) != -1:
	    # they want to substitute something in this command
	    pat = sub_regex.group(1)
	    repl = sub_regex.group(2)
	    input = regsub.gsub(pat,repl, input)
	return input

    def SendPlainInput(self, input):
	if input and data.currsession.connected:
	    if input[-1] != '\n' and input[-1] != '\r': 
		input = input + '\r'
	    mud.log(input)
	    mud.log(str(len(input)))
	    data.currsession.WriteTo(input)

	elif not data.numsessions:
	    ans = "no session active. " + \
		  "use the #session command to start one"
	    player.Putline(ans)
	    player.Prompt()

    # set the current session's data_hook, which better be a function
    def set_data_hook(self, func):
	if not type(func) == types.FunctionType:
	    raise 'set_data_hook', 'argument is not a function: %s'%func
	data.currsession.databuf.data_hook = func

    # speedwalk: send string of directions one letter at a time
    def DoSpeedWalk(self, input):
	# just send it if session has speedwalking off
	if not data.currsession.speedwalk:
	    self.SendPlainInput(input)
	    return

	send = ''
	i = 0
	num = ''
	n = len(input)
	# build a string called send which is purely letters.
	# convert stuff like 2e5s into eesssss.
	while i < n:
	    c = input[i]
	    if regex.match('[0-9]', c) != -1:
		# number: keep saving them till we see a char
		num = num + c
	    elif num:
		# tack num * c onto send
		send = send + c * string.atoi(num)
		num = ''
	    # just a character, tack it on
	    else:
		send = send + c
	    i = i + 1
	for s in send:
	    if s != '\n':
		self.SendPlainInput(s)
	
    # repeatedly execute a command
    def RepeatCommand(self, words):
	mud.log('repeat: received ' + string.join(words))
	if words:
	    if regex.match('^[0-9]+$', words[0]) != -1:
		num = string.atoi(words[0]) # number of repeats
		words = words[1:]
		for i in range(num):
		    self.HandleUserInput(string.join(words))
		return 1
	    else: 
		return 0

    # handle commands related to certain sessions
    def SessionCommand(self, words, ses):
	# change active session
	if ses and len(words) == 1:
	    if ses == 'all':
		player.Putline('all what?')
	    else:
		player.SetSes(ses)
		data.currsession.WriteTo('\n')
		data.currsession.WriteTo('\n')
		data.currsession.WriteTo('\n')
		return

	# command applies to all local sessions
	if ses == 'all':
	    if len(data.sessionlist) == 1:
		player.Putline('there aren\'t any sessions!')
		return
	    words = words[1:]
	    for s in data.sessionlist:
		if s is data.common:
		    continue
		oldses = data.currsession
		data.currsession = s
		self.HandleUserInput(string.join(words))
		data.currsession = oldses
	    return	    
	# command applies to a certain session
	else:
	    oldses = data.currsession
	    data.currsession = ses
	    words = words[1:]
	    self.HandleUserInput(string.join(words))
	    data.currsession = oldses
	    return


##################################################################
# Utility Functions
##################################################################

# run lyntin
def Run():
    cl = client()
    cl.CommandLine()
    player.Prompt()
    cl.Mainloop()

# return the whichth matched subgroup from regex( a compiled regex obj ),
# when matched against str
def BackRef(str, regex, which):
    if which == 0:
	return ''
    return str[regex.regs[which][0]:regex.regs[which][1]]

# return a 2-tuple representing the innards of two expressions within braces
def SplitBraced(str):
    nesting = 0
    one = ''
    two = ''
    parsed = ''
    mud.log(str)
    if string.find(str, '{') == -1:
	# unbraced, just chop it into two parts and return it
	sp = string.split(str)
	return sp[0], string.join(sp[1:])

    while str[0] == ' ':
	# trim leading whitespace
	str = str[1:]
    
    for c in str:
	if c == '{' and (not nesting) and parsed:
	    # we know one is unbraced, two braced
	    one, parsed = parsed, ''
	    nesting = nesting + 1
	    # chop off final whitespace from one
	    if one[-1] == ' ':
		one = one[:-1]
	elif c == '{':
	    if nesting:
		# preserve brace if we are alread nested
		parsed = parsed + c
	    nesting = nesting + 1
	elif c == '}':
	    if nesting > 1:
		# preserve brace if we are doubly(or more) nested
		parsed = parsed + c
	    nesting = nesting - 1 
	    if nesting == 0:
		if one:
		    two = parsed
		else:
		    one = parsed
	else:
	    parsed = parsed + c

    if nesting < 0:
	raise 'LTSyntaxError', 'unmatched braces'
    return one, two

# return the number of occurrences of pat in str
def CountRegex(pat, str):
    s = str[:]
    r = regex.compile(pat)
    count = 0
    while r.search(s) != -1:
	count = count + 1
	s = regsub.sub(pat, '', s)
    return count

# return a list of all the variables in a string
# no element will occur twice in the list
def StripVars(s):
    str = s[:]
    vars = []
    # regex to match nested variables within braces
    nested_var = regex.compile('{.*%%\([0-9]+\).*}')
    while nested_var.search(str) != -1:
	# found a nested variable: replace it with an unnested var
	pat = '%%'+nested_var.group(1)
	repl = '%'+nested_var.group(1)
	str = regsub.gsub(pat, repl, str)
    ret = str[:]
    while data.var_regex.search(str) != -1:
	var = data.var_regex.group(1)
	if var not in vars:
	    vars = vars + [var]
	str = regsub.sub(data.var_regex.givenpat, '', str)
    return vars, ret

# is input enclosed in braces?
def IsBrace(input):
    return input[0] == '{' and input[-1] == '}'

# check for sequencing, i.e. commands separated by ';'
# no choice but to do a grungy old parse
# on the way, we find syntax errors in lyntin commands
def IsSequence(input):
    if not input:
	return 0, ''
    ind = regex.search(';', input)
    # are we dealing with a lyntin command?
    # if not, then syntax errors don't matter
    if input[0] == '#':
	is_command = 1
    else:
	is_command = 0
    whether = 0
    seq = []
    parsed = ''
    if ind != -1:
	nesting = 0
	for c in input:
	    if nesting < 0 and is_command:
		raise 'LTSyntaxError', 'Unmatched Braces'
	    elif c == '{':
		nesting = nesting + 1
		parsed = parsed + c
	    elif c == '}':
		nesting = nesting -1
		parsed = parsed + c
	    elif c == ';' and nesting == 0:
		# this is a command sequence, unless the 
		# preceeding char was a backslash
		if not parsed:
		    # empty command
		    continue
		elif parsed[-1] == '\\':
		    parsed = parsed[:-1]
		    parsed = parsed + c
		else:
		    whether = 1
		    seq = seq + [parsed]
		    parsed = ''
	    else:
		parsed = parsed + c
	if nesting != 0 and is_command:
	    raise 'LTSyntaxError', 'Unmatched Braces'
    else:
	return 0, input

    if seq:
	# tack on final command
	seq = seq + [parsed]
	return whether, seq
    return whether, parsed


# is this a history command? (i.e. !n)
def IsHistory(input):
    return regex.search('^![0-9]*', input) != -1

# return the number a history command is referencing
# e.g !4 returns 4
def HistNumber(input):
    if regex.match('^!$', input) != -1 or \
       regex.match('^![^0-9]', input) != -1:
	return 0
    else:
	rx = regex.compile('!\([0-9]+\)')
	rx.search(input)
	return string.atoi(rx.group(1))

